package Ska::HashTable;

use warnings;
use PDL;
use PDL::NiceSlice;
use Data::Dumper;
use English;
use Carp;
use CFITSIO::Simple;
use RDB;
use POSIX qw(strtod);

require Exporter;

our @ISA = qw(Exporter);

our %EXPORT_TAGS = ( 'all' => [ qw() ] );
our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
our @EXPORT = qw();

our $VERSION = '0.04';
our $version = '$Id: HashTable.pm,v 1.11 2005/10/28 18:45:50 aldcroft Exp $';  # '
our $DEBUG = 0;
our %reserved = map {$_ => 1} qw(new add_col del_col read_data _convert_to_pdl is_pdl
				 print _make_accessor set_format _set_format reset rewind
				 row col row_values row_index);



# Preloaded methods go here.
#
# To Do:
#   - Make a uniform filtering interface, e.g. {filter => " ... ", cols => "...", rows=>"1-20"}
#     which works in any routine (including print, col, row)
#   - Option to set column ordering and format via dedicated set routine.
#     This should automatically occur as needed in print().


#****************************************************************************************
sub new {
#****************************************************************************************
    my $classname = shift;
    my $self = {};
    bless ($self);
    local $_;

    my $arg = shift;
    ($self->{data}, $self->{file_type}) = ref($arg) ? ($arg, 'none') : read_data($arg);

    my %arg = (accessor => 1,
	       @_);
    $self->{cols} = $arg{cols} || [ sort keys %{$self->{data}} ];
    $self->{fmt} = $arg{fmt} || {};

    foreach (keys %{$self->{data}}) {    
	$self->{is_pdl}->{$_} = UNIVERSAL::isa($self->{data}->{$_},'PDL');
    }

    my %n;
    foreach (values %{$self->{data}}) {
	my $n_rows = UNIVERSAL::isa($_,'PDL') ? ($_->dims)[-1] : scalar @{$_};
	$n{ $n_rows } = 1;
    }

    croak "HashTable::New: ERROR - inconsisent number of data elements\n"
      unless (scalar keys %n == 1);
    $self->{n_data} = (keys %n)[0];

    map { Ska::HashTable::_make_accessor($_) } @{$self->{cols}} if ($arg{accessor});

    return $self;
}


#****************************************************************************************
sub _make_accessor {
#
# Add an accessor function with the name of specified column, e.g.
# for a hashtable $ht with column 'column':
#  $column = $ht->column;     # Get the entire piddle or array ref
#  $value  = $ht->column(25); # Get element 25 of column
#
# This does NOT work as an lvalue
#****************************************************************************************
    my $col = $_[0];
    if ($reserved{$col}) {
	print STDERR "HashTable::_make_accessor - Warning: column $col conflicts with reserved keyword\n"
	  .          "     No accessor method created\n";
	return;
    }

    return if defined *{"Ska::HashTable::$col"};

    *{"Ska::HashTable::$col"} = sub {
	my $self = shift;
	if (@_) {
	    return $self->{is_pdl}->{$col} ? $self->{data}{$col}->at(@_) : $self->{data}{$col}->[$_[0]];
	} else {
	    return $self->{data}{$col};
	}
    };
}


#****************************************************************************************
sub add_col {
# Add specified columns to hashtable, with an optional hashref at
# the end which specifies the print format for columns
#
# add_col(col1 => $data1, ..., {col1 => $format1, ...})
#  $data1 is a PDL or array ref and $format1 is a printf string
#****************************************************************************************
    my $self = shift;
    my %fmt = (ref $_[-1] eq 'HASH') ? %{pop @_} : ();
    my %new = @_;
    my ($name, $data);
    local $_;

    while (my ($name, $data) = each %new) {
	my $n = UNIVERSAL::isa($data,'PDL') ? $data->nelem : scalar @{$data};
	croak "HashTable::Add: ERROR - inconsisent number of data elements for $name\n"
	  unless ($n == $self->{n_data});
	
	push @{$self->{cols}}, $name;
	$self->{data}{$name} = $data;
	$self->{is_pdl}{$name} = UNIVERSAL::isa($data,'PDL');
	$self->_set_format($name => $fmt{$name}) if defined $fmt{$name};
    }
}

#****************************************************************************************
sub del_col {
# Delete specified columns from hashtable
#****************************************************************************************
    my $self = shift;
    my $del;
    local $_;

    foreach $del (@_) {
	map { delete $self->{$_}{$del} } qw(data is_pdl fmt);
	@{$self->{cols}} = grep {$_ ne $del} @{$self->{cols}};
    }
}

#****************************************************************************************
sub read_data {
#****************************************************************************************
    # Guess the file type, either FITS, RDB, or white-space separated columns
    my $file = shift;
    my (%data, %datum);
    my $type;
    local $_;

    # If it looks like a FITS file, then use CFITSIO::Simple::fits_read_bintbl
    if ($file =~ /\.(fits|fts|fits\.gz|fts\.gz|ftz) (\[ [^\[\]]+ \])? \s* \Z/ix) {
	%data = fits_read_bintbl($file);
	$type = 'fits';
    } elsif ($file =~ /\.rdb \s* \Z/ix) {  # is it an RDB file?
	my $rdb = new RDB $file or die "Couldn't open $file\n";
	while ($rdb->read(\%datum)) {
	    map { push @{$data{$_}}, $datum{$_} } keys %datum;
	}
	_convert_to_pdl(\%data);
	$type = 'rdb';
    } else {			# Try whitespace-separated columns
	local @ARGV = ($file);
	my @cols;
	while (<>) {
	    next if (/\A (\s* \# | \s* \Z)/x);
	    (not @cols) and (@cols = split) and next;
	    my @datum = split;
	    croak "HashTable::read_data Error - invalid number of data columns in $file\n"
	      unless ($#datum = $#cols);
	    map { push @{$data{$cols[$_]}}, $datum[$_] } (0..$#cols);
	}
	_convert_to_pdl(\%data);
	$type = 'ascii';
    }	    
    return (\%data, $type);
}

#****************************************************************************************
sub _convert_to_pdl {
# Check each column (which must be an array-ref) for non-numeric entries
# by attempting to convert to a double and checking for problems.
# If none are found, convert to pdl
#****************************************************************************************
    my $data = shift;
    local $_;
    
    while (my ($name, $col) = each %{$data}) {
	my $numeric = 1;
	foreach (@{$col}) {
	    $! = 0;
	    my ($num, $n_unparsed) = POSIX::strtod($_);
	    if ($n_unparsed != 0 || $!) {
		$numeric = 0;
		last;
	    }
	}
	$data->{$name} = pdl $col if $numeric;
    }
}

#****************************************************************************************
sub is_pdl {
#****************************************************************************************
    my $self = shift;
    my $col  = shift;
    return $self->{is_pdl}->{$col};
}

#****************************************************************************************
sub print {
#****************************************************************************************
    my $self = shift;
    my @format = @_;
    my %data;
    local $_;

    # If format array ref not supplied, then build it with column names
    @format = @{$self->{cols}} unless (@format);
    my ($col,$fmt) = $self->_set_format(@format);

    # print header
    foreach (@{$col}) {
	my $len = ($fmt->{$_} =~ /%(-?\d*)/) ? $1 : '';
	printf "%${len}s", $_;
    }
    print "\n";

    # print data columns
    $self->rewind;
    while (%data = $self->row) {
	foreach (@{$col}) {
	    printf $fmt->{$_}, $data{$_};
	}
	print "\n";
    }
}

#**********************************************************************************
sub set_format {
#**********************************************************************************
    my $self = shift;
    $self->_set_format(@_, {force => 1});
}

#**********************************************************************************
sub _set_format {
#**********************************************************************************
    my $self = shift;
    my %par = (force => 0,
	       (ref $_[-1] eq 'HASH') ? %{pop @_} : ());
    my @format = @_;
    my %fmt;
    my @col;
    my $i;
    my $len;
    local $_;
    
    # Set up the column names, and formats where supplied
    foreach (@format) {
	if (/^\s*%/) {
	    my $col_name = $col[$#col];
	    $fmt{$col_name} = $_;
	    $self->{fmt}->{$col_name} = $_       # Store as default
	      unless (not $par{force} and $self->{fmt}->{$col_name}); 
	} else {
	    push @col, $_;
	    # Use default, which might get overridden by the next array element
	    $fmt{$_} = $self->{fmt}->{$_} if $self->{fmt}->{$_};  
	}
    }

    # Go through each column and make sure there is a format statement
    foreach (@col) {
	unless ($fmt{$_}) {
	    my $data_col = $self->col($_);
	    $len = length;	# start off with length of column name
	    my $is_pdl = $self->{is_pdl}->{$_};
	    for $i (0..$self->{n_data}-1) {
		my @fill = map {0} (0 .. scalar($data_col->dims) - 2) if $is_pdl;
		my $l = length ($is_pdl ? $data_col->at(@fill, $i) : $data_col->[$i]);
#		my $l = length ($is_pdl ? $data_col->at($i) : $data_col->[$i]);
		$len = $l if ($l > $len);
	    }
	    $len++;
	    $fmt{$_} = "%${len}s";
	    $self->{fmt}->{$_} = $fmt{$_} unless $self->{fmt}->{$_};  # Store as default
	}
    }

    # If forcing new format, then re-arrange column ordering as necessary by pushing
    # all the explicitly specified columns to the front (while retaining previous order)
    if ($par{force}) {
	my %col;
	map { $col{$_} = 1 } @col;
	my @new_cols = @col;
	map { push @new_cols, $_ unless $col{$_} } @{$self->{cols}};
	@{$self->{cols}} = @new_cols;
    }

    return \@col, \%fmt;
}

#****************************************************************************************
sub reset {
#****************************************************************************************
    my $self = shift;
    undef $self->{filter};
    undef $self->{row_list};
    undef $self->{row_index};
    undef $self->{fmt};
}

#****************************************************************************************
sub rewind {
#****************************************************************************************
    my $self = shift;
    $self->{row_index} = 0;
}

#****************************************************************************************
sub row {
#****************************************************************************************
    local $_;
    my $self = shift;
    my $filter = shift || "";
    
    return $self->row_values($filter) if (not wantarray);

    # If the filter doesn't match previous filter then reset filter
    if (not defined $self->{filter} or $filter ne $self->{filter}) {
	$self->{row_list} = $self->row_index($filter);
	$self->{row_index} = 0;
	$self->{filter} = $filter;
    }

    # If past the end of the filter list, return nothing
    return if $self->{row_index} >= $self->{row_list}->nelem;

    my $i = $self->{row_list}->at($self->{row_index});
    my %record;
    my ($name, $data);
    while (($name, $data) = each %{$self->{data}}) {
	my @fill = map {0} (0 .. scalar($data->dims) - 2) if $self->{is_pdl}->{$name};
	$record{$name} = $self->{is_pdl}->{$name} ? $data->at(@fill, $i) : $data->[$i];
    } 
    $self->{row_index}++;

    return %record;
}

#****************************************************************************************
sub col {
#****************************************************************************************
    my $self = shift;
    return $self->{data}->{ shift @_ };
}

#****************************************************************************************
sub row_values {
#****************************************************************************************
    my $self = shift;
    my ($out, $expr_ind, @expr_ind);
    my $p = $self->{data};
    local $_;

    # Use eval to evaluate the final expression and find the indices of selected elements
    $expr_ind =  $self->row_index(shift);
    @expr_ind = list $expr_ind;

    if ($expr_ind->nelem == 0) {   # no matches, return empty lists
	foreach (keys %{$p}) {
	    $out->{$_} = $self->{is_pdl}->{$_} ? pdl [] : [];
	}
    } else {
	foreach (keys %{$p}) {
	    # Depending on data type, use PDL NiceSlicing or Perl slicing
	    $out->{$_} = $self->{is_pdl}->{$_} ? $p->{$_}->($expr_ind) : [ @{$p->{$_}}[@expr_ind] ];
	}
    }
    return Ska::HashTable->new($out,
			       fmt => $self->{fmt},
			       cols => $self->{cols});
}

#****************************************************************************************
sub row_index {   
#****************************************************************************************
    my $self = shift;
    my $p = $self->{data};	# Input hash of piddles/array refs (for notational conven.)
    my $expr = shift;		# Filter expressions or index lists
    my ($out, $i, @ind, $filter, $expr_ind, @expr_ind);
    my $eval_expr = $expr;
    local $_;

    return pdl [] unless $self->{n_data} > 0;

    return sequence($self->{n_data}) unless $expr;

    while ($expr =~ /([^\(\)\&\|\s]+) \s* 
			    (\seq\s|\sne\s|==|=\~|!\~|>=|>|<|<=|!=) \s* 
			    ([^\(\)\&\|\s=<>!]+) /xg) {
	# Generate array of 0's and 1's corresponding to each component conditional
	# E.g. filter(\%aipr, "obsid == 123")  =>  ($aipr->{obsid} == 123)
	# This array has the same length as the input arrays
	my ($col, $expr) = ($1, "$2$3");
	my $eval = UNIVERSAL::isa($p->{$col},'PDL') ? "(\$p->{'$col'} $expr)" :
	  "[map {scalar(\$p->{'$col'}->[\$_] $expr)} (0..\$#{\$p->{'$col'}})]";
	print "Eval = $eval\n" if $DEBUG;
	$i = eval $eval;

	# Store list into an array of PDL indices
	push @ind, UNIVERSAL::isa($i,'PDL') ? $i : pdl $i;

	# Substitute the list of 0,1's into the full filter expression (which may 
	# include arbitrarily complex grouping parens and logicals)
	$eval_expr =~ s/$MATCH/\$ind\[$#ind\]/;
    }

    # Change logical && to multiply and || to addition (for lists of 0/1's)
    $eval_expr =~ s/&&/*/g;
    $eval_expr =~ s/\|\|/+/g;
    $eval_expr =  "which ($eval_expr)";

    # Use eval to evaluate the final expression and find the indices of selected elements
    $expr_ind =  eval $eval_expr;
    return $expr_ind;
}

1;
__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

Ska::Filter - Perl extension for blah blah blah

=head1 SYNOPSIS

  use Ska::Filter;
  blah blah blah

=head1 ABSTRACT

  This should be the abstract for Ska::Filter.
  The abstract is used when making PPD (Perl Package Description) files.
  If you don't want an ABSTRACT you should also edit Makefile.PL to
  remove the ABSTRACT_FROM option.

=head1 DESCRIPTION

Stub documentation for Ska::Filter.

=head2 EXPORT

None by default.



=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

If you have a mailing list set up for your module, mention it here.

If you have a web site set up for your module, mention it here.

=head1 AUTHOR

Tom Aldcroft, E<lt>aldcroft@localdomainE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2003 by Tom Aldcroft

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=cut
