#!/usr/bin/perl

# Install modules for yaxx

use warnings;
use strict;
use Cwd;
use Getopt::Long;
use File::Basename;
use File::Copy;

my %opt = ( force => 0,
            dryrun => 0,
            local => 1,
       );

GetOptions(\%opt,
           "prefix=s",
	   "dryrun!",
	   "force!",
	   "local!",
	   "exclude!",
	   );

# Run the sysarch sh script to determine the platform.  With the -perl switch
# it generates perl commands which we then eval to define the architecture.
my $sysarch_def = `script/sysarch -perl_hash`;
my %sysarch = eval "( $sysarch_def )";
my $ARCH = $sysarch{platform_os_generic};

my $YAXXDIST = cwd();
my $INSTALL = $opt{prefix} ? Cwd::abs_path($opt{prefix}) : "$YAXXDIST/$ARCH";
my $BUILD   = "$YAXXDIST/build/$ARCH";
my $PKGS     = "$YAXXDIST/pkgs";
my $PERL5LIB = "$INSTALL/lib/perl";
$| = 1;

$ENV{YAXX_DIR} = $INSTALL;
$ENV{PERL5LIB} = join(':', "$PERL5LIB", "$PERL5LIB/lib", $ENV{PERL5LIB} || () );
$ENV{CFITSIO} = $INSTALL;

# Create directories as needed
for my $dir ($INSTALL, $BUILD, $PERL5LIB) {
    run("mkdir -p $dir");
}

# Install yaxx source files
print "\nInstalling yaxx source files\n";
my $SRC = join(' ', qw(src/yaxx.cfg src/yaxx.pl src/yaxx src/lib
                       doc script resources slang Makefile background_model));
run("rsync -a $SRC $INSTALL/");

# Copy test suite 
run("rsync -a --exclude XMM Test $INSTALL/");

# Set the list of packages.  Those with two words are interpreted as
# code programs that get built with configure.  The first word is the
# tarball root name, the second is the directory that tarfile creates.
#
# The --NOTEST packages can fails in tests for various reasons even though
# the package is just fine.

my @packages =  ( 'cfitsio3060 cfitsio',
		    qw( 
		       IO::String
		       IO::All
		       Compress::Zlib
		       Archive::Zip
		       List::MoreUtils
		       IO::Tty
		       YAML
		       Tie::IxHash
		       Clone
		       Class::MakeMethods
		       Config::General
		       Hash::Merge
		       Expect      
		       Expect::Simple 
		       PDL
                       RDB
		       Params::Validate
		       Astro::FITS::CFITSIO
		       Astro::FITS::CFITSIO::CheckStatus
		       Astro::FITS::Header
		       CFITSIO::Simple
		       Astro::FITS::CFITSIO::Simple
                       Time::Out
		       Env::Path
		       Shell::GetEnv
		       Time-modules
		       Time::DayOfYear
		       Ska::Convert
		       Ska::HashTable
		       Ska::Message
		       Ska::Run
		       Decision::Depends
		       Chandra::Tools::dmcoords--NOTEST
		       Module::Build
		       version
		       Readonly
		       Sub::Uplevel
		       Test::Simple
		       Test::Exception
		       IO::stringy
		       Text::TabularDisplay
		       Text::RecordParser
		       Text::Glob
		       HTML::Tagset
		       HTML::Parser
		       HTML::TableExtract
		       File::MMagic
		       HTML::Table
		       Regexp::Common
		       Data::ParseTable
		       Chandra::Tools::ECF
		      ));

if ($opt{exclude}) {
    foreach my $exclude (@ARGV) {
	@packages = grep { not /^ $exclude /x } @packages;
    }
} else {
    @packages = @ARGV if @ARGV;
}

# Modules that require use Build.PL instead of Makefile.PL
my %use_Build_PL = map { $_ => 1 } qw(Module::Build
				      Text::RecordParser
				      version
				     );

# Try to 'use' these modules for specified packages.  The default is the package name
my %test_modules_for =  ('Time-modules' => [ qw(Time::JulianDay Time::ParseDate Time::CTime) ],
			 'IO::stringy'   => [ qw(IO::Stringy IO::AtomicFile IO::Scalar IO::Lines) ],
                         'Test::Simple' => [qw(Test::Simple Test::Builder::Tester)],
			);

foreach (@packages) {
    my @comp = split;
    (@comp == 1) ? perl_module_install(@comp) : code_install(@comp);
}

sub perl_module_install {
    my $module = shift;

    # if module name has '--NOTEST' at the end, then remove this and set flag so that
    # 'make test' is not run
    my $no_test = ($module =~ s/(..NOTEST)//g);

    # Get the package name 
    (my $pkg_name = $module) =~ s/::/-/g;

    print "\n**************** Build and install $module *****************\n\n";

    # First see if the module can be used
    # 
    print "Testing for module accessibility\n";
    my @test_modules = defined $test_modules_for{$module} ? @{$test_modules_for{$module}} : ($module);
    my $use_module_failed;
    foreach my $test_module (@test_modules) {
	print "Trying to use module $test_module ... ";
	$use_module_failed = system(qq(env PERL5LIB=$ENV{PERL5LIB} perl -e 'use $test_module' 2>&1 > /dev/null));
	print $use_module_failed ? "NOT OK\n" : "OK\n";
	last if $use_module_failed;
    }

    unless ($opt{local} or $opt{force}) {
	unless ($use_module_failed) {
	    print "  Already installed\n";
	    return;
	}
    }

    # Look for gunzipped tar files matching module name
    run_cd("$PKGS");
    my @files = glob("*.tar.gz");
    my @pkgs = grep /\A${pkg_name}-[^-]+\Z/, @files;
    die "Looking for the module ${pkg_name}-*.tar.gz in $PKGS, found '@pkgs' instead\n"
      unless (@pkgs == 1);
    (my $build = $pkgs[0]) =~ s/(_patch)?\.tar\.gz//;
	
    if (-e "$BUILD/$build/.installed") {
	if ($opt{force} or $use_module_failed) {
	    run("rm $BUILD/$build/.installed");
	} else {
	    print "  Already installed\n";
	    return;
	}
    }

    my $Makefile_PL = $use_Build_PL{$module} ? "Build.PL --install_base $PERL5LIB"
      : "Makefile.PL PREFIX=${PERL5LIB} LIB=${PERL5LIB}";
    my $make = $use_Build_PL{$module} ? "./Build" : "make";

    run_cd("$BUILD");
    run("gunzip --stdout $PKGS/$pkgs[0] | tar xvf -");
    run_cd("$build");
    run("perl $Makefile_PL");
    run("$make");
    run("$make test") unless $no_test;
    run("$make install");
    run("touch .installed");
}
    

sub code_install {
    my ($pkg, $build) = @_;
    print "\n**************** Build and install $build *****************\n\n";

    if (-e "$BUILD/$build/.installed") {
	if ($opt{force}) {
	    run("rm $BUILD/$build/.installed");
	} else {
	    print "  Already installed\n";
	    return;
	}
    }

    run_cd("$BUILD");
    run("gunzip --stdout $PKGS/${pkg}.tar.gz | tar xvf -");
    run_cd("$build");
    run("./configure --prefix=$INSTALL");
    run("make");
    run("make install");
    run("rm *.o");
    run("touch .installed");
}

sub run {
    my $cmd = shift;
    print "$cmd\n";
    return if $opt{dryrun};

    if (system($cmd)) {
	print "\n\n***** WARNING - '$cmd' appears to have failed ***\n\n";
	print " Options: (c)ontinue anyway  (q)uit : ";
	my $input = <STDIN>;
	die unless $input =~ /\A \s* c/ix;
	print "Continuing...\n\n";
    }
}

sub run_cd {
    my $dir = shift;
    my $cmd = "cd $dir";
    print "$cmd\n";
    return if $opt{dryrun};

    unless (chdir $dir) {
	print "\n\n***** ERROR - '$cmd' did not succeed ***\n\n";
	die;
    }
}

sub vers_num {
# Convert a string version to a numerical version, e.g.
# 3.2.53 -> 3.002053
    my $v = shift;
    my @v = split /[\.]/, $v;
    my $mult = 1;
    my $version = 0;
    local $_;
    foreach (@v) {
	$version += $_ * $mult;
	$mult /= 1000;
    }
    return $version;
}
