# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package Yaxx;

use strict;
use warnings;

# 'Local' packages
use Ska::Run qw(run);
use Ska::Convert qw(dec2hms);
use CFITSIO::Simple;
use Chandra::Tools::dmcoords;
use Chandra::Tools::ECF;
use Yaxx::ConfigFile;

# Available on CPAN
use HTML::Entities ();
use HTML::Table;
use IO::All;
use Regexp::Common;
use Decision::Depends;
use Data::Dumper;
use Data::ParseTable;
use POSIX qw(tmpnam strtod);
use Carp;
use File::Basename;
use File::Spec;
use File::Copy;
use File::Glob qw(:glob);
use Hash::Merge;
use Cwd;
use English;
use Storable qw(dclone);
use PDL;
use List::MoreUtils;
use Text::Glob;
use Regexp::Common;

use Class::MakeMethods::Standard::Hash ( 
  scalar => [ qw( 
		 allow_failed_fit 
		 asol_glob 
		 analysis_dir 
		 badpix_glob
		 bgd_ann_sep 
		 bgd_ann_wid
		 bgd_model_file
		 binned_method 
		 binned_stat
		 clean
		 clean_spec
		 common_model_defs
		 config_file 
		 mission
		 detector
		 environment
		 evt2_glob 
		 failed_at
		 fake_instances
		 fakeid
		 file_definition 
		 fit_plot
		 fit_rules
		 force
		 group_val
		 html_report_index
		 input_dir 
		 lock 
		 max_energy 
		 min_counts 
		 min_energy 
		 min_src_rad model 
		 formatting_rule
		 next_source
		 obsid 
		 postprocess_step
		 preclean
		 preprocess_step
		 prev_source
		 process 
		 process_step
		 program_dir
		 projection
		 psf_fraction 
		 psf_energy
		 redshift_column_name
		 source_data 
		 source_image
		 spec_chan_type 
		 srcid
		 src2_glob
		 src_reg
		 stopped_at
		 summary_header_key_col
		 task
		 thread
		 timeout
		 unbinned_method 
		 unbinned_stat 
		 uncertainty
		 use_psextract
		 use_environment
		 verbose_master_log
		 verbose_source_log
		 verbose_stdout
		 write_MDL_header_keys
		 log_time_format
		 output_dir
		 objlist						       
		 bkglist 
		 log_file
		 src_column_name
		 src_net_cts_name
		) ]
				       );

# Main package methods

our $FATAL_ERROR = <<'MESSAGE';

A fatal error has occurred.  If you believe this is not due to
faulty user inputs, please copy error message and the stack trace 
below and send to

  taldcroft@cfa.harvard.edu

MESSAGE

# our %formatting_rule;
our $NAN = (-1)**0.5;		# Generate a NaN 
our $DMcoord;
our $DMcoord_obsid = -1;
our %Bkg;			# Hash {obsid}{srcid} of bkg Yaxx objects
our $ECF;			# Chandra::Tools::ECF (Encircled Counts Fraction) object
our $Arcsec_per_pixel = 0.492;	# Arcsec per ACIS pixel
our %Message;

Hash::Merge::set_behavior( 'RIGHT_PRECEDENT' );

##****************************************************************************
sub new {
##****************************************************************************
    my $classname = shift;
    my $self = {};
    bless ($self, $classname);
    local $_;

    # Merge the default config, project config, and source info hashes into $par
    # using Hash::Merge routine with right precedence
    my $par = {};
    map { $par = Hash::Merge::merge($par, $_) } @_;

    # Bootstrap certain object data required for $self->file() to find other yaxx.cfg files
    foreach (qw (file_definition obsid detector ccdid srcid fakeid output_dir config_file)) {
	$self->$_($par->{$_}) if defined $par->{$_};
    }

#    # Merge config (yaxx.cfg) from obsid and source directory
#    foreach ( qw(obsid_dir src_dir) ) {
#	my $cfg_file = Yaxx::ConfigFile->new( name     => $self->config_file,
#					      path     => $self->file($_),
#					      required => 0,
#					      type     => $_,
#					    );
#	if ($cfg_file->parsed) {
#	    message(2, "Reading config file $_\n") ;
#	    $par = $cfg_file->merge($par);
#	}
#    }

    # Finally set all object data via accessor methods
    foreach (keys %{$par}) {
	eval { $self->$_( $par->{$_} ) };
	if ($@) {
	    message(0, "ERROR - user supplied config parameter '$_' is not allowed\n");
	    confess $Yaxx::FATAL_ERROR;		# So long
	}
    }

    return $self;
}

########################################################################################
sub listify {
########################################################################################
    if (defined $_[0]) {
	return ref($_[0]) eq 'ARRAY' ? @{$_[0]} : ($_[0]);
    } else {
	return ();
    }
}

##****************************************************************************
sub interpolate_struct {
# Use Data::Dumper and interpolate_string to interpolate a structure
##****************************************************************************
    my $self = shift;
    my $struct = shift;

    my $struct_dump;
    eval {
	$struct_dump = $self->interpolate_string( Data::Dumper->new([$struct])->Terse(1)->Dump );
    };
    if ($@) {
	message(1, "ERROR interpolate_struct - interpolation failed: $@\n");
	return;
    }
    my $struct_interp = eval $struct_dump;
    if ($@) {
	message(1, "ERROR - evaluation of Dump output for struct interpolation failed: $@\n");
	return;
    }
    return $struct_interp;
}

##****************************************************************************
sub interpolate_string {
#
#  Do interpolation of a config variable:
#    %VALUE{var, [options] }%   : yaxx variable 'var', possibly formatted
#    %FILE{type, [options] }%   : $self->file(type, [options])
#    %SPECIAL_INTERP%           : substitute with corresponding value passed to routine
# 
#  Options are comma separated name=value pairs.  Expected options are:
#    format   :   printf style formatting specification
#    detector :   Detector (e.g. for XMM this can be mos1, mos2 or pn)
#    fakeid   :   Typically for specifying fakeid=0 to force access to non-fake files
##****************************************************************************
    my $self = shift;
    my $str  = shift;
    my %opt = ( type => 'plain',   # Type of output: plain, latex, html
		@_,                # Possibly incldes 'special' substitutions
	      );

    my $interp_RE = qr/ % 
			([A-Z0-9_]*)
			($RE{balanced}{-parens=>'{}'})?
			% /x;

    # Keep pounding on string, finding occurences of %VALUE{ stuff }%
    # or likewise and substituting as needed.

    while ($str =~ /$interp_RE/) {
	my $match_str = $MATCH;	# Entire string that will be replaced
	my $interp_type = $1;	# FILE or VALUE or ...
	my $arg = $2;		# file type or variable to interpolate
	my $interp_val;		# Final interpolated value

	# If an argument e.g. {obsid} was given, clean it and interpolate
	if (defined $arg) {
	    # Recursively interpolate the argument string if req'd
	    $arg = $self->interpolate_string($arg, %opt) if $arg =~ /$interp_RE/;

	    # Clean argument and parse comma-separated file/variable name and interp. options
	    # (e.g. format, detector, fakeid).  (Clean 
	    $arg =~ s/ \A { \s* | \s* } \Z //xg;
	    $arg =~ s/ , \Z //xg; # Clean trailing commas because of bug in Text::RecordParser
	    my ($var_name, @interp_opts) = Text::RecordParser->new(data => $arg,
								field_separator=> ',',
								trim=>1,
							       )->fetchrow_array;
	    # Parse interpolation options like: format = '%05d' , fakeid=0
	    my %interp_opt;
	    foreach my $interp_opt (@interp_opts) {
		next unless (defined $interp_opt and $interp_opt =~ /\S/); # Need some non-whitespace stuff
		my @val = Text::RecordParser->new(data => $interp_opt,
						  field_separator=> '=',
						  trim=>1
						 )->fetchrow_array;

		# Do basic syntax check that option looks like name=value
		unless (@val==2 and defined $val[0] and defined $val[1]) {
		    confess "ERROR - interpolation option '$interp_opt' does not match name=value syntax\n";
		}

		# Strip outer quotes if needed, e.g. FORMAT=' %05d ' 
		$val[1] = $3 if ($val[1] =~ /\A $RE{delimited}{-keep}{-delim=>q{"}} \Z/x or
				 $val[1] =~ /\A $RE{delimited}{-keep}{-delim=>q{'}} \Z/x);
		$interp_opt{$val[0]} = $val[1];
	    }		

	    # Temporarily override specified yaxx variables.  The options FORMAT, ABSOLUTE,
	    # and CONTENT are upper-cased to hopefully avoid collisions with user source data columns.
	    # The code is a bit gruesome because I couldn't figure a better way to get
	    # the scope of the looped 'local' statement into this context
	    my $make_locals;
	    foreach (keys %interp_opt) {
		$make_locals .= "local \$self->{";
		$make_locals .= $self->can($_) ? "$_};" : "source_data}->{$_};" ;
	    }
	    $make_locals .= <<'SET_VALUES';
            foreach (keys %interp_opt) {
		($self->can($_) ? $self->{$_} : $self->{source_data}->{$_}) = $interp_opt{$_};
	    } ;
SET_VALUES

	    $interp_opt{type} = $opt{type};
	    if ($interp_type eq 'FILE') {
		eval $make_locals . '$interp_val = $self->file($var_name, \%interp_opt )';
		confess "ERROR - eval failed: $@" if $@;
	    } elsif ($interp_type =~ /^(VALUE|FORMAT)$/) {
		my $def_fmt = ($interp_type eq 'FORMAT');
		eval $make_locals . '$interp_val = $self->format_var($var_name, {DEFAULT_FMT=>$def_fmt, %interp_opt})';
		confess "ERROR - eval failed: $@" if $@;
		confess "interpolate_string - unexpected interpolation var '$var_name' found\n"
		  unless  defined $interp_val;
	    } elsif ($interp_type eq 'LABEL') {
		$interp_val = $self->get_formatting_rule($var_name)->{label} || '';
	    } else {
		confess "interpolate_string - unexpected interpolation type '$interp_type' found\n";
	    }
        # If no argument was given, it is a special subst, e.g. %HTML_FIT_TABLE% that
	# is supplied to routine.  Make sure we have something valid.  No formatting
        # is applied nor is the output type filter done
	} elsif (not defined ($interp_val = $opt{$interp_type})) {
	    confess "interpolate_string - unexpected interpolation type '$interp_type' found\n";
	}
	
	$str =~ s/$match_str/$interp_val/g; # replace multiple occurences if possible
    }

    return $str;
}

##****************************************************************************
sub next_process_step {
##****************************************************************************
    my $self = shift;
    my $pp = shift || '';  # 'pre' or 'post' or ''
    my $ps = $self->{"${pp}process_step"};
    
    # Relevant process step defined?
    return unless defined $ps;

    # Force into arrayref
    $ps = [ listify($ps) ];
    my $n = @{$ps};

    my $i = $self->{"curr_${pp}process_step"};
    my $next_name;
    if (defined $i) {
	$next_name = $ps->[$i]{next} || $ps->[++$i]{name};
    } else {
	$i = 0;
	$next_name = $ps->[$i]{name};
    }

    return if ($i >= $n);

    if (defined $next_name) {
	return if ($next_name eq 'END_OF_PROCESSING');
	($i) = grep { defined $ps->[$_]{name} && $ps->[$_]{name} eq $next_name } reverse (0..$n-1);
	if (not defined $i) {
	    message(1, "ERROR - could not find next process step '$next_name'\n");
	    confess "\n";
	}
    }

    $ps->[$i]{type} = $pp;
    $self->{"curr_${pp}process_step"} = $i;
    return $ps->[$i];
}


##****************************************************************************
sub get_var {
# Get a yaxx variable
##****************************************************************************
    my $self = shift;
    my $var  = shift;
    my $value = $self->can($var) ? $self->$var : $self->source_data->{$var};
}

##****************************************************************************
sub set_var {
# Set a yaxx variable
##****************************************************************************
    my $self = shift;
    my $opt  = shift;
    my @var  = listify($opt->{var}) or confess "Need a 'var' parameter specifying variable name";
    my $value = $opt->{value};
    confess "Need a 'value' parameter specifying variable value" unless defined $value;
    chomp $value if $opt->{chomp};
    foreach my $var (@var) {
	$self->can($var) ? $self->$var($value)
	                 : ($self->source_data->{$var} = $value);
	message(4, "Setting '$var' = '$value'\n");
    }
    return 1;
}

##****************************************************************************
sub unique_identifier {
# Make a string that uniquely identifies the source
##****************************************************************************
    my $self = shift;
    my @id;

    # Report on each identity attribute that has an accessor method
#    foreach my $attr (qw(obsid ccdid srcid)) {
#	push @id, $attr . '=' . $self->$attr if $self->can($attr);
#    }
    foreach my $attr (qw(obsid ccdid srcid)) {
	push @id, $attr . '=' . $self->{$attr} if defined $self->{$attr};
    }

    # Fakeid is a special case.  Only include if fakeid is true (indicating fake source)
    push @id, 'fakeid' . '=' . $self->fakeid if ($self->can('fakeid') and $self->fakeid);
    

    # Make the first character uppercase
    map { substr($_, 0, 1) = uc substr($_, 0, 1) } @id;

    return join('  ', @id);
}

##****************************************************************************
sub message {
##****************************************************************************
    # Get some options off an optional hashref at the end
    my %more_opt = (ref $_[-1] eq 'HASH') ? %{pop @_} : ();

    # If first arg is a Ska::Message object then initialize package vars
    if (UNIVERSAL::isa($_[0], "Ska::Message")) {
	my $msg = shift;

	if ($more_opt{close}) {
	    delete $Message{$msg};
	    return;
	}

	my $verbose_stdout = shift if @_;
	my $verbose_logfile = shift if @_;
	$Message{$msg} = { ska_msg_obj => $msg,
			   verbose_stdout => $verbose_stdout,
			   verbose_logfile => $verbose_logfile
			 };
	return;
    }

    croak "ERROR - no message objects defined\n" unless keys %Message;

    croak "Need a message!" unless defined $_[0];

    my $verbose = (@_ and $_[0] =~ /\A \s* \d \s* \Z/x) ? shift : -1;
    foreach my $msg_printing_object (values %Message) {
	my %opt;
	# Turn off printing to stdout or file if verbosity threshold for the given message text
	# is greater than the respective message-printing-object verbosity
	$opt{stdout} = undef if $verbose > $msg_printing_object->{verbose_stdout};
	$opt{file}   = undef if $verbose > $msg_printing_object->{verbose_logfile};
	%opt = (%opt, %more_opt);

	# Actually write out message to stdout, logfile as specified.  
	$msg_printing_object->{ska_msg_obj}->message(@_, \%opt);
    }
}

############################################################################3
sub run_tool {
############################################################################3
    my $cmd = shift;
    my $cmd_pid;

    # First pull any hash ref off the end.  This specifies parameters for
    # the run_tool subroutine (e.g. 'paste' and 'timeout')
    my %par = (loud => 0);
    if (ref $_[-1] eq 'HASH') {
	%par = (%par, %{pop @_});
    }
    local %ENV = CXC::Envs::env($par{ENV}) if $par{ENV};
    
    my %toolpar = @_;		# Get the parameters for the tool to be run

    my ($cmd_root) = split ' ', $cmd;
    my $sep = $par{paste} ? " " : "\n    ";
    system("punlearn $cmd_root") if (keys %toolpar);

    message(3, "Running $cmd$sep") if $par{loud};

    foreach (keys %toolpar) {
	my $arg = $par{unix_params} ? " -$_ $toolpar{$_}" : " $_=$toolpar{$_}";
	$cmd .= $arg;
	message(4, "$arg$sep") if $par{loud};
    }
    message(3, "\n") if $par{loud};

    # Run within eval block to be able to set a local alarm to time out if
    # command doesn't finish in in $par{timeout} seconds
    # See perldoc -f alarm
    eval {
	local $SIG{ALRM} = sub { message(1, "ERROR - $cmd_root command timed out ".localtime()."\n");
				 die "alarm\n";
			       }; # NB: \n required
	alarm $par{timeout} if $par{timeout};
	
	$cmd_pid = open CMD, "$cmd 2>&1 |"
	  or die "run_tool: ERROR - Could not start $cmd_root command\n";
	while (<CMD>) {
	    message(4, $_) if $par{loud};
	    $ {$par{out}} .= $_ if $par{out}; 
	}
	close CMD;

	alarm 0;
    };

    if ($@) {
	die unless $@ eq "alarm\n"; # propagate unexpected errors
	# timed out.  Find all child processes and kill them one at a time

	message(1, "WARNING - Caught the timeout alarm at " . localtime() . "\nKilling processes...\n");

	my $ps_cmd = '/bin/ps -a -o "pid pgid args"';
	my $xfit_gid = getpgrp;
	my @ps = `$ps_cmd`;
	foreach (@ps) {
	    my ($pid, $gid, $cmd) = split;
	    if ($gid =~ /^\d+$/ and $gid == $xfit_gid and $pid != $$) {
		kill 9 => $pid;
		message(1, "WARNING - Killed pid gid = $_");
		sleep 2;
	    }
	}
	sleep 5;
	close CMD;

	return;
    }

    1;
}
1;

