# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package Yaxx;
use strict;
use warnings;

##****************************************************************************
sub run_extract {
##****************************************************************************
# 
#  Run psextract to extract PI file
#
##****************************************************************************
    my $self = shift;
    local $_;
    my %file;

    chdir $self->file('src_dir') if $self->file('src_dir');

    # Check for required input files
    my @depend   = qw(src bkg evt2);
    push @depend, 'asol' if $self->{use_psextract};
    push @depend, 'badpix' if (-e $self->file('badpix')); # Optional bad pixel file
    for (@depend) {
	unless (-e ($file{$_} = $self->file($_))) {
	    message(1, "ERROR - missing $_ file $file{$_}\n");
	    return;
	}
    }

    # Set up to check dependencies and return success if already met
    my @target   = map { $self->file($_) } qw(pi rmf arf bg_pi asphist);
    my @test_dep = (-target => \@target, -depend => [ values %file ]);

    return 1 unless test_dep(@test_dep);

    # Delete any existing psextract output files
    my @psfiles = map { $self->file($_) } qw(pi rmf arf bg_pi bg_rmf bg_arf asphist);
    map {unlink if -e} @psfiles; # Clean all psextract files first

    # Define filtered source and background files
    my $energy = "energy=".$self->min_energy.":".$self->max_energy;
    my $src_evts_energy = qq{"$file{evt2}\[sky=region($file{src})][$energy]"};
    my $src_evts 	= qq{"$file{evt2}\[sky=region($file{src})]"};
    my $bkg_evts 	= qq{"$file{evt2}\[sky=region($file{bkg})]"};
	
    # Make sure there are some counts in source region
    if (get_evt_counts($src_evts_energy) < 1) {
	message(1, "ERROR - no counts in source region\n");
	return;
    }

    # If a bad pixel file is present, run acis_set_ardlib
    if ($file{badpix}) {	# (Already tested for existence earlier)
	run_tool("acis_set_ardlib $file{badpix}", { loud => 1 });
    }

    # Now actually run psextract to extract the spectrum
    run_tool("punlearn dmgroup");
    my @extract = $self->{use_psextract} ? 
      ("psextract",
       events   => $src_evts,
       bgevents => $bkg_evts,
       root     => $self->file('root'),
       asol     => $file{asol}, 
       bgasol   => "",
       ptype    => $self->spec_chan_type,
       gtype    => "NONE",
       gspec    => 0,
      )
	:
	  ("specextract",
	   infile   => $src_evts,
	   outroot  => $self->file('root'),
	   bkgfile  => $bkg_evts,
	   ptype    => $self->spec_chan_type,
	   grouptype => "NONE",
	  );
    
    my $run_tool_out;
    run_tool(@extract,
	    verbose  => 2,
	    clobber  => "yes",
	    {
	     timeout => $self->timeout, 
	     paste => 1,
	     loud  => 1,
	    }
	   );

    # Make sure all dependencies are now satisfied
    if (test_dep(@test_dep)) {
	message(1, "ERROR - did not create all files in psextract\n");
	return;
    }

    1;
}

#********************************************************************************
sub get_obs_info_from_pi {
#********************************************************************************
    my $self = shift;
    my $ok;

    return unless (-e $self->file('pi') and -e $self->file('bg_pi'));

    my %rmf = fits_read_bintbl($self->file('rmf'), {extname => 'EBOUNDS'});
    $ok = PDL::which($rmf{e_min} > $self->min_energy_kev);
    my $min_bin = ($ok->nelem > 0) ? $ok->at(0)+1 : 1;

    $ok = PDL::which($rmf{e_max} < $self->max_energy_kev);
    my $max_bin = ($ok->nelem > 0) ? $ok->at(-1)+1 : $rmf{e_max}->nelem;

    my %counts;
    for my $type (qw(pi bg_pi)) {
	run_tool('dmstat',
		 infile => $self->file($type) . qq{'[channel=${min_bin}:${max_bin}][cols counts]'},
		 { loud => 0,
		 },
		);
	
	run_tool('pget dmstat out_sum',
		 { out => \$counts{$type},
		   loud => 0,
		 }
		);
	$counts{$type} =~ s/\s//g;
    }


    my $hdr = fits_read_hdr($self->file('pi'), 'spectrum');
    my $hdr_bkg = fits_read_hdr($self->file('bg_pi'), 'spectrum');
    $self->counts( $counts{pi} - $counts{bg_pi} * $hdr->{BACKSCAL} / $hdr_bkg->{BACKSCAL} );
    $self->exposure( $hdr->{EXPOSURE} );
}

1;
