# --8<--8<--8<--8<--
#
# Copyright (C) 2006 Smithsonian Astrophysical Observatory
#
# This file is part of yaxx
#
# yaxx is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# yaxx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the 
#       Free Software Foundation, Inc. 
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package Yaxx;
use strict;
use warnings;

##****************************************************************************
sub get_fit_bkg_model_cmd {
##****************************************************************************
    my $self = shift;
    my $file;
    if (defined $self->bgd_model_file) {
	$file = io($self->bgd_model_file);
	$file = io->catfile($self->analysis_dir, $self->bgd_model_file) unless $file->is_absolute;
    } else {
	my $hdr = fits_read_hdr($self->file('arf'), 'SPECRESP');
	my @bgd_hdr_vals = ($hdr->{TELESCOP}, $hdr->{INSTRUME});
	push @bgd_hdr_vals, $hdr->{DETNAM} if defined $hdr->{DETNAM};
	$file = io->join($self->file('bkg_mdl_dir', {ABSOLUTE=>1}),
			 join("_", @bgd_hdr_vals) . '.in'
			)->name;
	$file =~ s/\s//g;
    }
	
    unless (-e $file) {
	message(1, "ERROR - Cannot find background model file $file\n");
	return;
    }
    return "use $file";
}

##****************************************************************************
sub fit_spectral_models {
##****************************************************************************
    my $self = shift;
    local $_;
    chdir $self->file('src_dir') if $self->file('src_dir');

    # Customize each of the models to insert appropriate redshift, Gal NH etc
    foreach (values %{$self->model}) {
	$_ = $self->interpolate_string($_);
    }	

    # Parse a table that looks like:

    # fit_rules  <<END_FIT_RULES_TABLE
    # # model_name      Fit file	        condition
    # # ---------   -------------------    -------------
    #   pl              fit_ungrouped	      counts <= 80
    #   pl_fix_abs      fit_ungrouped	      counts <= 80
    #   pl              fit_grouped	      counts >  80
    #   pl_fix_abs      fit_grouped	      counts >  80
    #   pl_abs          fit_grouped	      counts >  200
     #END_FIT_RULES_TABLE

    my @fit_rules;
    foreach (split("\n", $self->fit_rules)) {
	# Strip out comments and make sure something is left
	s/#.*//;
	next unless /\S/;
	my @fields = split;
	if (@fields < 3) {
	    message(1, "ERROR - 'fit_rules' table needs 3 columns: model_name group condition\n");
	    confess $Yaxx::FATAL_ERROR;
	}
	push @fit_rules, { model_name        => shift(@fields),
			   fit_template_file => io->join($self->file('resources_fit_dir'), shift(@fields))->name,
			   condition         => join(q{ }, @fields),
			 };

	# Do some error checking
	my $fs = $fit_rules[-1];
	unless (defined $fs->{model_name} and defined $self->model->{$fs->{model_name}}) {
	    message(1, "ERROR - model name in yaxx.cfg 'fit_rules' not defined\n");
	    message(1, "        Line is '$_'\n");
	    confess $Yaxx::FATAL_ERROR;
	}
	unless (-e $fs->{fit_template_file}) {
	    message(1, "ERROR - Cannot find fit template file $fs->{fit_template_file}\n");
	    message(1, "        Line is '$_'\n");
	    confess $Yaxx::FATAL_ERROR;
	}
	if (not defined $fs->{condition}) {
	    $fs->{condition} = 1; # Default to always true
	}
    }

    # Do the actual fitting!  If fit failed, then either bail on obsid (default)
    # or push on with other model fits
    foreach my $fit_rule (@fit_rules) {
	next unless $self->eval_fit_condition($fit_rule->{condition});
	if (not $self->fit_ciao($fit_rule)) {
	    message(1, "ERROR - Unsuccessful fit\n");
	    return unless $self->allow_failed_fit;
	}
    }

    return 1;
}

##****************************************************************************
sub eval_fit_condition {
#
# Parse condition string and substitute as needed so that it can be eval'ed
# E.g. change condition => 'counts > 30 and class eq "quasar"'
# to '$self->counts > 30 and $self->source_data->{class} eq "quasar"'
##****************************************************************************
    my $self = shift;
    my @user_accessible_attributes = qw(exposure 
					obsid ccdid srcid
					counts
					X Y
					ra dec rad
					object
					gal_nh
				       );
    my $rule = shift;
    
    my $condition;
    while ($rule =~ /\G ([^'"]+ | $RE{quoted} ) /xg) {
	my $stuff = $1;
	if ($stuff =~ /$RE{quoted}/) {
	    $condition .= $stuff;
	} else {
	    my @words = split /\b/, $stuff;
	    foreach my $word (@words) {
		if (grep { $word eq $_ } @user_accessible_attributes) {
		    $word = qq/\$self->$word/;
		} elsif (defined $self->source_data->{$word}) {
		    $word = qq/\$self->source_data->{'$word'}/;
		}
	    }
	    $condition .= join('', @words);
	}
    }

    my $ok = eval $condition;
    if ($@) {
	message(1, "ERROR - Could not evaluate fit rule '$rule':\n");
	message(1, "$@\n");
	confess $Yaxx::FATAL_ERROR;
    }

    return $ok;
}

##****************************************************************************
sub fit_ciao {
##****************************************************************************
    my $self = shift;
    my $fit_rule = shift;
    local $_;

    my $mdl_name = $fit_rule->{model_name};
    my $cmd_file = "$mdl_name.in";
    my $out_file = "$mdl_name.out";

    # Assemble dependencies and check
    my @target = ("$mdl_name.mdl", "$mdl_name.in");
    my @depend = map { $self->file($_) } qw(pi bg_pi rmf arf);
    my @test_dep = (-target => \@target, -depend => \@depend );

    return 1 unless test_dep(@test_dep);
    unless ($self->model->{$mdl_name}) {
	message(1, "ERROR - Unknown model name '$mdl_name' specified\n");
	return;
    }

    # Clean any existing sherpa fit files for this model
    map { unlink } glob("$mdl_name.*"); 

    # Get the appropriate sherpa fit commands based on the fit_template and then
    # interpolate
    my $cmds < io($fit_rule->{fit_template_file});
    $cmds = $self->interpolate_string($cmds,
				      MODEL_NAME        => $mdl_name,
				      SOURCE_MODEL      => $self->{model}{$mdl_name},
				      TITLE_MODEL_NAME  => latex_verbatim($mdl_name),
				      CREATION_DATE     => scalar(localtime),
				      FIT_PLOT_WIDTH_PIXELS => $self->{fit_plot}{width_pixels},
				     );
    $cmds .= "\n";

    eval { $cmds > io($cmd_file) };
    croak "ERROR - could not open $cmd_file for sherpa commands: \n $@" if ($@);

    # Do the actual sherpa model fit
    message(3, "Fitting model $mdl_name\n");
    my $sherpa_out;
    return unless run_tool("sherpa --batch $cmd_file", { out => \$sherpa_out,
							 timeout => $self->timeout,
							 loud => 1,
						       } );

    # Parse results of projection and/or uncertainty
    my %proj_result = parse_limits_table($sherpa_out, 'sherpa.proj') if $self->projection;
    my %unc_result = parse_limits_table($sherpa_out, 'sherpa.unc') if $self->uncertainty;
    return unless %proj_result or %unc_result;

    # Add projection/uncertainty results into mdl file if method was done and results obtained
    update_mdl_file($mdl_name, \%proj_result, 'proj') if %proj_result;
    update_mdl_file($mdl_name, \%unc_result, 'unc') if %unc_result;

    if (test_dep(@test_dep)) {
	message(1, "ERROR - fit failed to produce all files\n");
	return;
    }

    return 1;
}

##****************************************************************************
sub update_mdl_file {
#
# Put upper and lower limit values (from projection/uncertainty) in MDL file
# as new columns
##****************************************************************************
    # MDL file name, Results of projection/uncertainty, and type ('proj' or 'unc')
    my ($mdl_name, $proj, $type) = @_; 
    my ($model_name, $model, $comp_name, $comp);
    my @parname;
    my %upper;
    my %lower;
    local $_;

# Straight from CIAO::Sherpa::Parse
    while (($model_name, $model) = each %{$proj}) {
	while (($comp_name, $comp) = each %{$model}) {
	    $upper{"$model_name.$comp_name"} = $comp->{upper};
	    $lower{"$model_name.$comp_name"} = $comp->{lower};
	}
    }

    # Read in existing MDL file for this $mdl_name, then get parameter
    # upper and lower limits where they exist.  Put in a hash and call fits_write_bintbl

    my %mdl = fits_read_bintbl("$mdl_name.mdl[MDL_Models]", 'parname');
    my @lower_lims = map { $lower{$_} || 0.0 } @{$mdl{parname}};
    my @upper_lims = map { $upper{$_} || 0.0 } @{$mdl{parname}};
    my %out = ("${type}_lower" => \@lower_lims,
	       "${type}_upper" => \@upper_lims);
    fits_write_bintbl("$mdl_name.$type", %out);

    # Paste the new columns into the MDL_Models extension of the MDL file, then rebuild
    # the entire file (with both blocks) using dmcopy and dmappend

    my $tmp = POSIX::tmpnam();
    my $tmp2 = POSIX::tmpnam();
    run_tool("dmpaste",
	     infile => "$mdl_name.mdl[MDL_Models]",
	     pastefile => "$mdl_name.$type",
	     outfile => $tmp,
	    {loud => 0});
    run_tool("dmcopy", infile => "$mdl_name.mdl[MDL_Data]", outfile => $tmp2, {loud => 0});
    run_tool("dmappend", infile => $tmp, outfile => $tmp2, {loud => 0});
    move $tmp2, "$mdl_name.mdl";
    unlink $tmp;

    message(3, "Successfully updated $mdl_name.mdl file\n");
}

##****************************************************************************
sub parse_limits_table {
##****************************************************************************
  my ( $PS )  = shift;
  my $command = shift;
  my $Null_Lim = qr/--+/;

  # next, we look for the results.
  $PS =~ /Computed for\s+$command\.(\S+)[ \t]*=[ \t]*(.*)\n/igc    # CIAO 3.0.1
    or do { message(1, "ERROR - Couldn't find $command results\n");
	    return};

  my $method_var = $1;
  my $method_val = $2;

  my %result;

  # next three lines are like this:
  #      --------------------------------------------------------
  #      Parameter Name      Best-Fit Lower Bound     Upper Bound
  #      --------------------------------------------------------

  $PS =~ /\G\s+-+\n/gc && 
    $PS =~ /\G\s+Parameter Name.*\n/gc &&
      $PS =~ /\G\s+-+\n/gc
	or do {message(1, "ERROR - out of sync for $command command; expecting Results header");
	       return};

  while ( $PS =~ /\G
                  [ \t]*(\S+)\.(\S+)
                  [ \t]+($RE{num}{real})
                  [ \t]+($RE{num}{real}|$Null_Lim)
                  [ \t]+($RE{num}{real}|$Null_Lim)
                  .*\n
                 /xgc )
  {
    # can't assign it in the regexp match above, else \G doesn't get set right
    my ( $model, $param, $best, $lower, $upper ) = ( $1, $2, $3, $4, $5 );
    $upper = $Yaxx::NAN if ($upper =~ /$Null_Lim/);
    $lower = $Yaxx::NAN if ($lower  =~ /$Null_Lim/);

    $result{$model}{$param}{best}  = $best;
    $result{$model}{$param}{lower} = $lower;
    $result{$model}{$param}{upper} = $upper;
  }

  return %result;
}

1;

